import { addDeletedToSelect, stripDeletedFieldFromResults, } from "./utils/nestedReads";
import { filterSoftDeletedResults, shouldFilterDeletedFromReadResult, } from "./utils/resultFiltering";
/* Delete middleware */
function createDeleteParams(params, { field, createValue }) {
    if (!params.model ||
        // do nothing for delete: false
        (typeof params.args === "boolean" && !params.args)) {
        return params;
    }
    if (typeof params.args === "boolean") {
        return {
            ...params,
            action: "update",
            args: {
                __passUpdateThrough: true,
                [field]: createValue(true),
            },
        };
    }
    return {
        ...params,
        action: "update",
        args: {
            where: params.args.where || params.args,
            data: {
                [field]: createValue(true),
            },
        },
    };
}
export function createDeleteMiddleware(config) {
    return function deleteMiddleware(params, next) {
        return next(createDeleteParams(params, config));
    };
}
/* DeleteMany middleware */
function createDeleteManyParams(params, config) {
    if (!params.model)
        return params;
    const where = params.args.where || params.args;
    return {
        ...params,
        action: "updateMany",
        args: {
            where: {
                ...where,
                [config.field]: config.createValue(false),
            },
            data: {
                [config.field]: config.createValue(true),
            },
        },
    };
}
export function createDeleteManyMiddleware(config) {
    return function deleteManyMiddleware(params, next) {
        return next(createDeleteManyParams(params, config));
    };
}
/* Update middleware */
function createUpdateParams(params, config) {
    var _a, _b;
    if (((_a = params.scope) === null || _a === void 0 ? void 0 : _a.relations) &&
        !params.scope.relations.to.isList &&
        !config.allowToOneUpdates &&
        !params.args.__passUpdateThrough) {
        throw new Error(`prisma-soft-delete-middleware: update of model "${params.model}" through "${(_b = params.scope) === null || _b === void 0 ? void 0 : _b.parentParams.model}.${params.scope.relations.to.name}" found. Updates of soft deleted models through a toOne relation is not supported as it is possible to update a soft deleted record.`);
    }
    // remove __passUpdateThrough from args
    if (params.args.__passUpdateThrough) {
        delete params.args.__passUpdateThrough;
    }
    return params;
}
export function createUpdateMiddleware(config) {
    return function updateMiddleware(params, next) {
        return next(createUpdateParams(params, config));
    };
}
/* UpdateMany middleware */
function createUpdateManyParams(params, config) {
    return {
        ...params,
        args: {
            ...params.args,
            where: {
                ...params.args.where,
                // allow overriding the deleted field in where
                [config.field]: params.args.where[config.field] || config.createValue(false),
            },
        },
    };
}
export function createUpdateManyMiddleware(config) {
    return function updateManyMiddleware(params, next) {
        return next(createUpdateManyParams(params, config));
    };
}
/* Upsert middleware */
export function createUpsertMiddleware(_) {
    return function upsertMiddleware(params, next) {
        var _a, _b;
        if (((_a = params.scope) === null || _a === void 0 ? void 0 : _a.relations) && !params.scope.relations.to.isList) {
            throw new Error(`prisma-soft-delete-middleware: upsert of model "${params.model}" through "${(_b = params.scope) === null || _b === void 0 ? void 0 : _b.parentParams.model}.${params.scope.relations.to.name}" found. Upserts of soft deleted models through a toOne relation is not supported as it is possible to update a soft deleted record.`);
        }
        return next(params);
    };
}
/* FindUnique middleware */
function createFindUniqueParams(params, config) {
    return {
        ...params,
        action: "findFirst",
        args: {
            ...params.args,
            where: {
                ...params.args.where,
                [config.field]: config.createValue(false),
            },
        },
    };
}
export function createFindUniqueMiddleware(config) {
    return function findUniqueMiddleware(params, next) {
        return next(createFindUniqueParams(params, config));
    };
}
/* FindFirst middleware */
function createFindFirstParams(params, config) {
    return {
        ...params,
        action: "findFirst",
        args: {
            ...params.args,
            where: {
                ...params.args.where,
                // allow overriding the deleted field in where
                [config.field]: params.args.where[config.field] || config.createValue(false),
            },
        },
    };
}
export function createFindFirstMiddleware(config) {
    return function findFirst(params, next) {
        return next(createFindFirstParams(params, config));
    };
}
/* FindMany middleware */
function createFindManyParams(params, config) {
    return {
        ...params,
        action: "findMany",
        args: {
            ...params.args,
            where: {
                ...params.args.where,
                // allow overriding the deleted field in where
                [config.field]: params.args.where[config.field] || config.createValue(false),
            },
        },
    };
}
export function createFindManyMiddleware(config) {
    return function findManyMiddleware(params, next) {
        return next(createFindManyParams(params, config));
    };
}
/* Count middlware */
function createCountParams(params, config) {
    const args = params.args || {};
    const where = args.where || {};
    return {
        ...params,
        args: {
            ...args,
            where: {
                ...where,
                // allow overriding the deleted field in where
                [config.field]: where[config.field] || config.createValue(false),
            },
        },
    };
}
export function createCountMiddleware(config) {
    return function countMiddleware(params, next) {
        return next(createCountParams(params, config));
    };
}
/* Aggregate middleware */
function createAggregateParams(params, config) {
    const args = params.args || {};
    const where = args.where || {};
    return {
        ...params,
        args: {
            ...args,
            where: {
                ...where,
                // allow overriding the deleted field in where
                [config.field]: where[config.field] || config.createValue(false),
            },
        },
    };
}
export function createAggregateMiddleware(config) {
    return function aggregateMiddleware(params, next) {
        return next(createAggregateParams(params, config));
    };
}
/* Where middleware */
function createWhereParams(params, config) {
    var _a;
    // customise list queries with every modifier unless the deleted field is set
    if (((_a = params.scope) === null || _a === void 0 ? void 0 : _a.modifier) === "every" && !params.args[config.field]) {
        return {
            ...params,
            args: {
                OR: [
                    { [config.field]: { not: config.createValue(false) } },
                    params.args,
                ],
            },
        };
    }
    return {
        ...params,
        args: {
            ...params.args,
            [config.field]: params.args[config.field] || config.createValue(false),
        },
    };
}
export function createWhereMiddleware(config) {
    return function whereMiddleware(params, next) {
        if (!params.scope)
            return next(params);
        return next(createWhereParams(params, config));
    };
}
/* Include middleware */
function createIncludeParams(params, config) {
    var _a, _b, _c;
    // includes of toOne relation cannot filter deleted records using params
    // instead ensure that the deleted field is selected and filter the results
    if (((_b = (_a = params.scope) === null || _a === void 0 ? void 0 : _a.relations) === null || _b === void 0 ? void 0 : _b.to.isList) === false) {
        if (params.args.select && !params.args.select[config.field]) {
            return addDeletedToSelect(params, config);
        }
        return params;
    }
    return {
        ...params,
        args: {
            ...params.args,
            where: {
                ...params.args.where,
                // allow overriding the deleted field in where
                [config.field]: ((_c = params.args.where) === null || _c === void 0 ? void 0 : _c[config.field]) || config.createValue(false),
            },
        },
    };
}
export function createIncludeMiddleware(config) {
    return async function includeMiddleware(params, next) {
        var _a;
        const updatedParams = createIncludeParams(params, config);
        const deletedFieldAdded = typeof updatedParams.args === "object" &&
            ((_a = updatedParams.args) === null || _a === void 0 ? void 0 : _a.__deletedFieldAdded);
        if (deletedFieldAdded) {
            delete updatedParams.args.__deletedFieldAdded;
        }
        const result = await next(updatedParams);
        if (shouldFilterDeletedFromReadResult(params, config)) {
            const filteredResults = filterSoftDeletedResults(result, config);
            if (deletedFieldAdded) {
                stripDeletedFieldFromResults(filteredResults, config);
            }
            return filteredResults;
        }
        return result;
    };
}
/* Select middleware */
function createSelectParams(params, config) {
    var _a, _b, _c, _d;
    // selects in includes are handled by createIncludeParams
    if (((_a = params.scope) === null || _a === void 0 ? void 0 : _a.parentParams.action) === "include") {
        return params;
    }
    // selects of toOne relation cannot filter deleted records using params
    if (((_c = (_b = params.scope) === null || _b === void 0 ? void 0 : _b.relations) === null || _c === void 0 ? void 0 : _c.to.isList) === false) {
        if (params.args.select && !params.args.select[config.field]) {
            return addDeletedToSelect(params, config);
        }
        return params;
    }
    return {
        ...params,
        args: {
            ...params.args,
            where: {
                ...params.args.where,
                // allow overriding the deleted field in where
                [config.field]: ((_d = params.args.where) === null || _d === void 0 ? void 0 : _d[config.field]) || config.createValue(false),
            },
        },
    };
}
export function createSelectMiddleware(config) {
    return async function selectMiddleware(params, next) {
        var _a;
        const updatedParams = createSelectParams(params, config);
        const deletedFieldAdded = typeof updatedParams.args === "object" &&
            ((_a = updatedParams.args) === null || _a === void 0 ? void 0 : _a.__deletedFieldAdded);
        if (deletedFieldAdded) {
            delete updatedParams.args.__deletedFieldAdded;
        }
        const result = await next(updatedParams);
        if (shouldFilterDeletedFromReadResult(params, config)) {
            const filteredResults = filterSoftDeletedResults(result, config);
            if (deletedFieldAdded) {
                stripDeletedFieldFromResults(filteredResults, config);
            }
            return filteredResults;
        }
        return result;
    };
}
